
// Take care of the deletion of a AIDA::Loader
// (in particular if allocated on the stack).
// Be sure to no more use code related to the Loader
// after the closing of it (done by the ~AIDA::Loader).

// Take care of the C functions declared to ::atexit in a Loader.
// On some system (RedHat-6.x) the dlclose does not call them !
// Instead they are called at the (final) exit. Then crashes can
// happen if someone dlclose the Loader before calling exit (or returning
// from the main).

// this :
#include "Loader.h"

#include <stdlib.h>

#ifdef WIN32
#include <windows.h>
#endif

#if defined(__alpha) || defined(__linux) || defined(__sun)
#define DLOPEN
#endif

#if defined(__APPLE__)
#define NSOPEN
#endif

#ifdef DLOPEN
#include <dlfcn.h>
#endif

#include <iostream>

#ifdef NSOPEN
#include <vector>
static std::vector<std::string> nspaths(); 
static void* nsopen(const std::string&);
static void* nssym(void*,const std::string&);
static bool nsclose(void*);
#endif

//////////////////////////////////////////////////////////////////////////////
AIDA::Loader::Loader(
 const std::string& aName
,bool aQuiet
,bool aVerbose
)
:fModule(0)
,fName(aName)
,fVerbose(aVerbose)
//////////////////////////////////////////////////////////////////////////////
// aName is the full name (path + libname) or could be simply
// the DLL name if access path are specified.
// Access path are :
// - PATH on Windows.
// - LD_LIBRARY_PATH on Linuxes.
// - DYLD_LIBRARY_PATH on MacOSX.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!open()) {
    if(!aQuiet) {
      std::cerr << "AIDA::Loader::Loader :" 
                << " can't load " << fName << " library." 
                << std::endl;
      std::cerr << "AIDA::Loader::Loader :" 
                << " error  : " << fError 
                << std::endl;
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
AIDA::Loader::~Loader(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if (fModule) {
    if(fVerbose) {
      std::cout << "AIDA::Loader::~Loader :" 
                << " close \"" << fName << "\"..." 
                << std::endl;
    }
#ifdef DLOPEN
    if(::dlclose(fModule)) {
      std::cerr << "AIDA::Loader::~Loader :" 
                << " error  : " << ::dlerror() 
                << std::endl;
    }
#endif
#ifdef NSOPEN
    nsclose(fModule);
#endif
#ifdef WIN32
    ::FreeLibrary((HMODULE)fModule);
#endif
    if(fVerbose) {
      std::cout << "AIDA::Loader::~Loader :" 
                << " close \"" << fName << "\" done." 
                << std::endl;
    }
  }
  fModule = 0;
}
//////////////////////////////////////////////////////////////////////////////
bool AIDA::Loader::open(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fError = "";
  if(fVerbose) {
    std::cout << "AIDA::Loader::open :" 
              << " open \"" << fName 
              << std::endl;
  }
#ifdef DLOPEN
  int flags = RTLD_NOW | RTLD_GLOBAL;
  fModule = ::dlopen(fName.c_str(),flags);
  if(!fModule) fError = ::dlerror();
#endif
#ifdef NSOPEN
  fModule = nsopen(fName);
  if(!fModule) {
    std::vector<std::string> paths = nspaths(); 
    for(unsigned int index=0;index<paths.size();index++) {
      std::string dll = paths[index]+"/" + fName + ".bundle";
      fModule = nsopen(dll);
      if(fModule) break;
    }
  }
#endif
#ifdef WIN32
  if(fVerbose) {
    std::cout << "AIDA::Loader::open :" 
              << " open \"" << fName << "\"..." 
              << std::endl;
  }
  fModule = ::LoadLibrary(fName.c_str());
#endif
  return fModule ? true : false;
}
//////////////////////////////////////////////////////////////////////////////
bool AIDA::Loader::isOpened(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fModule?true:false;
}
//////////////////////////////////////////////////////////////////////////////
const std::string& AIDA::Loader::fileName(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fName;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::Procedure AIDA::Loader::find(
 const std::string& aName
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fModule) return 0;
  Procedure p = 0;
#ifdef DLOPEN
  const char* cerror;
  p = (Procedure)::dlsym(fModule,(char*)aName.c_str());
  if ((cerror=dlerror())!=NULL) { //error : undefined symbol.
    std::string serror = cerror;
    std::string s = "_";
    s += aName;
    p = (Procedure)::dlsym(fModule,(char*)s.c_str());
    if ((cerror=dlerror())!=NULL) { //error : undefined symbol.
      std::cerr << "AIDA::Loader::findProcedure :" 
                << " problem when searching \"" << aName << "\""
                << " in \"" << fName << "\"."
                << " dlerror  : " << serror 
                << std::endl;
    } 
  }
#endif
#ifdef NSOPEN
  p = (Procedure)nssym(fModule,fName);
  if(!p) {
    std::string name = "_" + aName;
    p = (Procedure)nssym(fModule,name);
    if (!p) {
      std::cerr << "AIDA::Loader::findProcedure :" 
                << " problem when searching \"" << aName
                << "\" in " << fName << "."
		<< std::endl;
    } 
  }
#endif
#ifdef WIN32
  p = (Procedure)::GetProcAddress((HMODULE)fModule,aName.c_str());
  if (!p) {
    std::string s = "_";
    s += aName;
    p = (Procedure)::GetProcAddress((HMODULE)fModule,s.c_str());
    if (!p) {
      std::cerr << "AIDA::Loader::findProcedure :"
                << " can't load " << aName << " procedure." 
                << std::endl;
    }
  }
#endif
  return p;
}


#ifdef NSOPEN
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> Lib_smanip_words(
 const std::string& aString
,const std::string& aLimiter 
,bool aTakeEmpty = false
)
//////////////////////////////////////////////////////////////////////////////
//  If aLimiter is for exa "|" and for "xxx||xxx" :
//  - aTakeEmpty false : {"xxx","xxx"} will be created (and NOT {"xxx","","xxx"}).
//  - aTakeEmpty true : {"xxx","","xxx"} will be created.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> words;
  if(aString=="") return words;
  std::string::size_type lim = (aTakeEmpty?0:1);
  if(aLimiter=="") {
    words.push_back(aString);
  } else {
    std::string::size_type l = aString.length();
    std::string::size_type llimiter = aLimiter.length();
    std::string::size_type pos = 0;
    while(1) {
      std::string::size_type index = aString.find(aLimiter,pos);
      if(index==std::string::npos){ // Last word.
        if((l-pos)>=lim) words.push_back(aString.substr(pos,l-pos));
        break;
      } else {
        //     abcxxxef
        //     0  3  67
        if((index-pos)>=lim) words.push_back(aString.substr(pos,index-pos));
        pos = index + llimiter;
      }
    }
  }
  return words;
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> nspaths(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const char* env = ::getenv("DYLD_LIBRARY_PATH");
  std::string paths = (env==NULL ? std::string("") : std::string(env));
  return Lib_smanip_words(paths,":");
}
#include <mach-o/dyld.h>
//////////////////////////////////////////////////////////////////////////////
void* nsopen(
 const std::string& aDLL
) 
//////////////////////////////////////////////////////////////////////////////
//Foro infos : Darwin> man NSModule, NSObjectFileImage
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  NSObjectFileImage ofi = 0;
  NSObjectFileImageReturnCode code = 
    NSCreateObjectFileImageFromFile(aDLL.c_str(),&ofi);
  if(code!=NSObjectFileImageSuccess) return 0;
  NSModule module = NSLinkModule(ofi,aDLL.c_str(),FALSE);
  NSDestroyObjectFileImage(ofi);
  return module;
}
//////////////////////////////////////////////////////////////////////////////
void* nssym(
 void* aModule
,const std::string& aName
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  NSModule module = (NSModule)aModule;
  NSSymbol symbol = NSLookupSymbolInModule(module,aName.c_str());
  if(symbol==NULL) return 0;
  return NSAddressOfSymbol(symbol);
}
//////////////////////////////////////////////////////////////////////////////
bool nsclose(
 void* aModule
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  NSModule module = (NSModule)aModule;
  NSUnLinkModule(module,FALSE);
  return true;
}
#endif


//////////////////////////////////////////////////////////////////////////////
AIDA::IAnalysisFactory* AIDA::Loader::findAIDA(
 const std::string& aEntry
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string entry = aEntry;
  if(entry=="") entry = fName + "_createAnalysisFactory";

  typedef AIDA::IAnalysisFactory*(*CreateAnalysisFactoryFunction)();
  CreateAnalysisFactoryFunction func = 
   (CreateAnalysisFactoryFunction)find(entry);
  if(!func) {
    std::cout << "AIDA::Loader::findAIDA :" 
              << " " << entry << " not found in dll " << fName << "."
              << std::endl;
    return 0;
  }
  AIDA::IAnalysisFactory* af = (AIDA::IAnalysisFactory*)(*func)();
  if(!af) {
    std::cout << "AIDA::Loader::findAIDA :" 
              << " can't create an AIDA::IAnalysisFactory" 
              << " from dll " << fName << "."
              << std::endl;
    return 0;
  }
  return af;
}
